#include "stdafx.h"

int consoleBufferMaxSize = 16384;
std::string consoleBuffer;

void PrintHelp()
{
	printf("Usage: remote.exe -c <command>\r\n");
	printf("Example: remote.exe -c \"hlds.exe -console -game valve +ip 127.0.0.1 +port 27015 +map crossfire\"\r\n");
}

void ErrorMessage(char *str)
{
	LPVOID msg;
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR)&msg, 0, NULL);
	printf("%s: %s\n", str, msg);
	LocalFree(msg);
	_getch();
}

void StoreConsoleData(HANDLE hOutput, char *data)
{
	DWORD bytesWritten;
	WriteFile(hOutput, data, strlen(data), &bytesWritten, 0);

	int dataLen = strlen(data);
	if (dataLen > consoleBufferMaxSize)
	{
		consoleBuffer.assign(data + dataLen - consoleBufferMaxSize);
	}
	else
	{
		consoleBuffer.append(data);
		int size = consoleBuffer.size();
		if (size > consoleBufferMaxSize)
			consoleBuffer.erase(0, size - consoleBufferMaxSize);
	}
}

void WriteStringToConsoleInput(HANDLE hInput, char *stringToWrie)
{
	char *pos, c;
	char keyCode;
	INPUT_RECORD ir;
	DWORD numberOfEventsWritten;

	pos = stringToWrie;
	c = *pos;
	while (c != 0)
	{
		if (c == '\n')
			c = '\r';
		keyCode = toupper(c);

		ir.EventType = KEY_EVENT;
		ir.Event.KeyEvent.bKeyDown = TRUE;
		ir.Event.KeyEvent.wRepeatCount = 1;
		ir.Event.KeyEvent.wVirtualKeyCode = keyCode;
		ir.Event.KeyEvent.wVirtualScanCode = MapVirtualKey(keyCode, MAPVK_VK_TO_VSC);
		ir.Event.KeyEvent.uChar.UnicodeChar = (WCHAR)c;
		ir.Event.KeyEvent.dwControlKeyState = isupper(c) != 0 ? 0x80 : 0;

		if (!WriteConsoleInput(hInput, &ir, 1u, (LPDWORD)&numberOfEventsWritten))
			ErrorMessage("WriteConsoleInput");

		ir.Event.KeyEvent.bKeyDown = FALSE;

		if (!WriteConsoleInput(hInput, &ir, 1u, (LPDWORD)&numberOfEventsWritten))
			ErrorMessage("WriteConsoleInput");

		pos++;
		c = *pos;
	}
}

bool Start(const char *command)
{
	char buf[1024];

	STARTUPINFO si;
	SECURITY_ATTRIBUTES sa;
	SECURITY_DESCRIPTOR sd;
	PROCESS_INFORMATION pi;

	HANDLE hOutput, hInput;
	HANDLE stdout_read = NULL, stdout_write = NULL;

	if (IsWindowsXPOrGreater())
	{
		InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION);
		SetSecurityDescriptorDacl(&sd, true, NULL, false);
		sa.lpSecurityDescriptor = &sd;
	}
	else
	{
		sa.lpSecurityDescriptor = NULL;
	}

	sa.nLength = sizeof(SECURITY_ATTRIBUTES);
	sa.bInheritHandle = true;

	hInput = GetStdHandle(STD_INPUT_HANDLE);
	hOutput = GetStdHandle(STD_OUTPUT_HANDLE);

	if (!CreatePipe(&stdout_read, &stdout_write, &sa, 0))
	{
		ErrorMessage("CreatePipe");
		return false;
	}

	GetStartupInfo(&si);
	si.dwFlags = STARTF_USESTDHANDLES;
	si.wShowWindow = SW_HIDE | SW_MINIMIZE;
	si.hStdInput = hInput;
	si.hStdOutput = stdout_write;
	si.hStdError = stdout_write;

	char directory[MAX_PATH];
	strncpy(directory, command, sizeof(directory) - 1);
	directory[sizeof(directory) - 1] = 0;
	char *args = PathGetArgs(directory);
	*args = 0;
	char *pos1 = strrchr(directory, '\\');
	char *pos2 = strrchr(directory, '/');
	if (pos1 != NULL && pos1 > pos2)
		*pos1 = 0;
	else if (pos2 != NULL && pos2 > pos1)
		*pos2 = 0;

	if (!CreateProcess(NULL, (char *)command, NULL, NULL, TRUE, CREATE_NEW_CONSOLE, NULL, directory, &si, &pi))
	{
		ErrorMessage("CreateProcess");
		return false;
	}

	if (!FreeConsole())
	{
		ErrorMessage("FreeConsole");
		return false;
	}

	// Wait a bit for the process start
	Sleep(100);

	if (!AttachConsole(pi.dwProcessId))
	{
		ErrorMessage("AttachConsole");
		return false;
	}

	// Can check that we are both in the same boat
	//DWORD pl[10];
	//DWORD plCount = 10;
	//plCount = GetConsoleProcessList(pl, plCount);
	//if (plCount != 2)
	//{
	//	ErrorMessage("GetConsoleProcessList");
	//	return false;
	//}

	// Hide console window?
	//ShowWindow(GetConsoleWindow(), SW_HIDE);


	unsigned long exit = 0;
	unsigned long bytesRead, bytesAvailable;
	long bytesToRead;
	bool once = false;

	// Main loop
	while (true)
	{
		Sleep(10);
		GetExitCodeProcess(pi.hProcess, &exit);
		if (exit != STILL_ACTIVE)
			break;

		// Check if there is something to read from stdout
		PeekNamedPipe(stdout_read, buf, 1, &bytesRead, &bytesAvailable, NULL);
		if (bytesRead != 0)
		{
			bytesToRead = bytesAvailable;
			while (bytesToRead > 0)
			{
				ReadFile(stdout_read, buf, sizeof(buf) - 1, &bytesRead, NULL);
				buf[bytesRead] = 0;
				bytesToRead -= bytesRead;
				StoreConsoleData(hOutput, buf);
			}
		}

		if (!once && consoleBuffer.find("VAC secure mode is activated") != std::string::npos)
		{
			once = true;
			WriteStringToConsoleInput(hInput, "version\n");
			Sleep(10);
			WriteStringToConsoleInput(hInput, "status\n");
		}
	}

	return true;
}

int main(int argc, char* argv[])
{
	for (int i = 0; i < argc; i++)
	{
		char *param = argv[i];
		if (*param == '-' || *param == '/')
		{
			switch (param[1])
			{
			case 'c':
				if (i == argc - 1)
				{
					printf("Argument \"command\" is required.");
					exit(1);
				}
				if (!Start(argv[i + 1]))
					return 1;
				break;
			case '?':
				break;
				PrintHelp();
			}
		}
	}

	return 0;
}
